@@ -13,6 +13,7 @@ from .amsexceptions import (AmsServiceException, AmsConnectionException,
 from .amsmsg import AmsMessage
 from .amstopic import AmsTopic
 from .amssubscription import AmsSubscription
+from .amsuser import AmsUser
 
 try:
     from collections import OrderedDict
@@ -29,70 +30,149 @@ class AmsHttpRequests(object):
        library. service error handling is implemented according to HTTP
        status codes returned by service and the balancer.
     """
-    def __init__(self, endpoint):
+
+    def __init__(self, endpoint, authn_port, token="", cert="", key=""):
         self.endpoint = endpoint
+        self.authn_port = authn_port
+        self.token = token
+
         # Create route list
-        self.routes = {"topic_list": ["get", "https://{0}/v1/projects/{2}/topics?key={1}"],
-                       "topic_get": ["get", "https://{0}/v1/projects/{2}/topics/{3}?key={1}"],
-                       "topic_publish": ["post", "https://{0}/v1/projects/{2}/topics/{3}:publish?key={1}"],
-                       "topic_create": ["put", "https://{0}/v1/projects/{2}/topics/{3}?key={1}"],
-                       "topic_delete": ["delete", "https://{0}/v1/projects/{2}/topics/{3}?key={1}"],
-                       "topic_getacl": ["get", "https://{0}/v1/projects/{2}/topics/{3}:acl?key={1}"],
-                       "topic_modifyacl": ["post", "https://{0}/v1/projects/{2}/topics/{3}:modifyAcl?key={1}"],
-                       "sub_create": ["put", "https://{0}/v1/projects/{2}/subscriptions/{4}?key={1}"],
-                       "sub_delete": ["delete", "https://{0}/v1/projects/{2}/subscriptions/{4}?key={1}"],
-                       "sub_list": ["get", "https://{0}/v1/projects/{2}/subscriptions?key={1}"],
-                       "sub_get": ["get", "https://{0}/v1/projects/{2}/subscriptions/{4}?key={1}"],
-                       "sub_pull": ["post", "https://{0}/v1/projects/{2}/subscriptions/{4}:pull?key={1}"],
-                       "sub_ack": ["post", "https://{0}/v1/projects/{2}/subscriptions/{4}:acknowledge?key={1}"],
-                       "sub_pushconfig": ["post", "https://{0}/v1/projects/{2}/subscriptions/{4}:modifyPushConfig?key={1}"],
-                       "sub_getacl": ["get", "https://{0}/v1/projects/{2}/subscriptions/{3}:acl?key={1}"],
-                       "sub_modifyacl": ["post", "https://{0}/v1/projects/{2}/subscriptions/{3}:modifyAcl?key={1}"],
-                       "sub_offsets": ["get", "https://{0}/v1/projects/{2}/subscriptions/{3}:offsets?key={1}"],
-                       "sub_mod_offset": ["post", "https://{0}/v1/projects/{2}/subscriptions/{3}:modifyOffset?key={1}"],
-                       "auth_x509": ["get", "https://{0}:{1}/v1/service-types/ams/hosts/{0}:authx509"],
-                       "sub_timeToOffset": ["get", "https://{0}/v1/projects/{2}/subscriptions/{3}:timeToOffset?key={1}&time={4}"]
-                       }
+        self.routes = {
+            # topic api calls
+            "topic_list": ["get", "https://{0}/v1/projects/{1}/topics"],
+            "topic_get": ["get", "https://{0}/v1/projects/{1}/topics/{2}"],
+            "topic_publish": ["post", "https://{0}/v1/projects/{1}/topics/{2}:publish"],
+            "topic_create": ["put", "https://{0}/v1/projects/{1}/topics/{2}"],
+            "topic_delete": ["delete", "https://{0}/v1/projects/{1}/topics/{2}"],
+            "topic_getacl": ["get", "https://{0}/v1/projects/{1}/topics/{2}:acl"],
+            "topic_modifyacl": ["post", "https://{0}/v1/projects/{1}/topics/{2}:modifyAcl"],
+
+            # subscription api calls
+            "sub_create": ["put", "https://{0}/v1/projects/{1}/subscriptions/{2}"],
+            "sub_delete": ["delete", "https://{0}/v1/projects/{1}/subscriptions/{2}"],
+            "sub_list": ["get", "https://{0}/v1/projects/{1}/subscriptions"],
+            "sub_get": ["get", "https://{0}/v1/projects/{1}/subscriptions/{2}"],
+            "sub_pull": ["post", "https://{0}/v1/projects/{1}/subscriptions/{2}:pull"],
+            "sub_ack": ["post", "https://{0}/v1/projects/{1}/subscriptions/{2}:acknowledge"],
+            "sub_pushconfig": ["post", "https://{0}/v1/projects/{1}/subscriptions/{2}:modifyPushConfig"],
+            "sub_getacl": ["get", "https://{0}/v1/projects/{1}/subscriptions/{2}:acl"],
+            "sub_modifyacl": ["post", "https://{0}/v1/projects/{1}/subscriptions/{2}:modifyAcl"],
+            "sub_offsets": ["get", "https://{0}/v1/projects/{1}/subscriptions/{2}:offsets"],
+            "sub_mod_offset": ["post", "https://{0}/v1/projects/{1}/subscriptions/{2}:modifyOffset"],
+            "sub_timeToOffset": ["get", "https://{0}/v1/projects/{1}/subscriptions/{2}:timeToOffset?time={3}"],
+
+            # user api calls
+            "user_create": ["post", "https://{0}/v1/users/{1}"],
+            "user_get": ["get", "https://{0}/v1/users/{1}"],
+
+            # project api calls
+            "project_add_member": ["post", "https://{0}/v1/projects/{1}/members/{2}:add"],
+            "project_get_member": ["get", "https://{0}/v1/projects/{1}/members/{2}"],
+
+            "auth_x509": ["get", "https://{0}:{1}/v1/service-types/ams/hosts/{0}:authx509"],
+        }
 
         # HTTP error status codes returned by AMS according to:
         # http://argoeu.github.io/messaging/v1/api_errors/
-        self.ams_errors_route = {"topic_create": ["put", set([409, 401, 403])],
-                                 "topic_list": ["get", set([400, 401, 403,
-                                                            404])],
-                                 "topic_delete": ["delete", set([401, 403,
-                                                                 404])],
-                                 "sub_create": ["put", set([400, 409, 408, 401,
-                                                            403])],
-                                 "sub_ack": ["post", set([408, 400, 401, 403,
-                                                          404])],
-                                 "topic_get": ["get", set([404, 401, 403])],
-                                 "topic_modifyacl": ["post", set([400, 401,
-                                                                  403, 404])],
-                                 "sub_get": ["get", set([404, 401, 403])],
-                                 "topic_publish": ["post", set([413, 401,
-                                                                403])],
-                                 "sub_mod_offset": ["post", set([400, 401, 403,
-                                                                 404])],
-                                 "sub_pushconfig": ["post", set([400, 401, 403,
-                                                                 404])],
-                                 "auth_x509": ["post", set([400, 401, 403,
-                                                            404])],
-                                 "sub_pull": ["post", set([400, 401, 403,
-                                                           404])],
-                                 "sub_timeToOffset": ["get", set([400, 401,
-                                                                  403, 404,
-                                                                  409])]}
+        self.ams_errors_route = {
+            "topic_create": ["put", set([409, 401, 403])],
+            "topic_list": ["get", set([400, 401, 403, 404])],
+            "topic_delete": ["delete", set([401, 403, 404])],
+            "topic_get": ["get", set([404, 401, 403])],
+            "topic_modifyacl": ["post", set([400, 401, 403, 404])],
+            "topic_publish": ["post", set([413, 401, 403])],
+
+            "sub_create": ["put", set([400, 409, 408, 401, 403])],
+            "sub_get": ["get", set([404, 401, 403])],
+            "sub_mod_offset": ["post", set([400, 401, 403, 404])],
+            "sub_ack": ["post", set([408, 400, 401, 403, 404])],
+            "sub_pushconfig": ["post", set([400, 401, 403, 404])],
+            "sub_pull": ["post", set([400, 401, 403, 404])],
+            "sub_timeToOffset": ["get", set([400, 401, 403, 404, 409])],
+
+            "user_create": ["post", set([400, 401, 403, 404, 409])],
+            "user_get": ["post", set([400, 401, 403, 404])],
+
+            "project_add_member": ["post", set([400, 401, 403, 404, 409])],
+            "project_get_member": ["get", set([400, 401, 403, 404])],
+
+            "auth_x509": ["post", set([400, 401, 403, 404])]}
+
         # https://cbonte.github.io/haproxy-dconv/1.8/configuration.html#1.3
         self.balancer_errors_route = {"sub_ack": ["post", set([500, 502, 503, 504])],
                                       "sub_pull": ["post", set([500, 502, 503, 504])],
                                       "topic_publish": ["post", set([500, 502, 503, 504])]}
 
+        # determine the token to be used
+        self.assign_token(token, cert, key)
+
+    def assign_token(self, token, cert, key):
+        """Assign a token to the ams object
+
+           Args:
+               token(str): a valid ams token
+               cert(str): a path to a valid certificate file
+               key(str): a path to the associated key file for the provided certificate
+        """
+
+        # check if a token has been provided
+        if token != "":
+            return
+
+        try:
+            # otherwise use the provided certificate to retrieve it
+            self.token = self.auth_via_cert(cert, key)
+        except AmsServiceException as e:
+            # if the request send to authn didn't contain an x509 cert, that means that there was also no token provided
+            # when initializing the ArgoMessagingService object, since we only try to authenticate through authn
+            # when no token was provided
+
+            if e.msg == 'While trying the [auth_x509]: No certificate provided.':
+                refined_msg = "No certificate provided. No token provided."
+                errormsg = self._error_dict(refined_msg, e.code)
+                raise AmsServiceException(json=errormsg, request="auth_x509")
+            raise e
+
+    def auth_via_cert(self, cert, key, **reqkwargs):
+        """Retrieve an ams token based on the provided certificate
+
+            Args:
+                cert(str): a path to a valid certificate file
+                key(str): a path to the associated key file for the provided certificate
+
+           Kwargs:
+               reqkwargs: keyword argument that will be passed to underlying
+                          python-requests library call.
+        """
+        if cert == "" and key == "":
+            errord = self._error_dict("No certificate provided.", 400)
+            raise AmsServiceException(json=errord, request="auth_x509")
+
+        # create the certificate tuple needed by the requests library
+        reqkwargs = {"cert": (cert, key)}
+
+        route = self.routes["auth_x509"]
+
+        # Compose url
+        url = route[1].format(self.endpoint, self.authn_port)
+        method = getattr(self, 'do_{0}'.format(route[0]))
+
+        try:
+            r = method(url, "auth_x509", **reqkwargs)
+            # if the `token` field was not found in the response, raise an error
+            if "token" not in r:
+                errord = self._error_dict("Token was not found in the response. Response: " + str(r), 500)
+                raise AmsServiceException(json=errord, request="auth_x509")
+            return r["token"]
+        except (AmsServiceException, AmsConnectionException) as e:
+            raise e
+
     def _error_dict(self, response_content, status):
         error_dict = dict()
 
         try:
-            if (response_content and sys.version_info < (3, 6, ) and
-                isinstance(response_content, bytes)):
+            if (response_content and sys.version_info < (3, 6,) and
+                    isinstance(response_content, bytes)):
                 response_content = response_content.decode()
             error_dict = json.loads(response_content) if response_content else {}
         except ValueError:
@@ -155,9 +235,13 @@ class AmsHttpRequests(object):
                         saved_exp = e
                         time.sleep(sleep_secs)
                         if timeout:
-                            log.warning('Backoff retry #{0} after {1} seconds, connection timeout set to {2} seconds - {3}: {4}'.format(i, sleep_secs, timeout, self.endpoint, e))
+                            log.warning(
+                                'Backoff retry #{0} after {1} seconds, connection timeout set to {2} seconds - {3}: {4}'.format(
+                                    i, sleep_secs, timeout, self.endpoint, e))
                         else:
-                            log.warning('Backoff retry #{0} after {1} seconds - {2}: {3}'.format(i, sleep_secs, self.endpoint, e))
+                            log.warning(
+                                'Backoff retry #{0} after {1} seconds - {2}: {3}'.format(i, sleep_secs, self.endpoint,
+                                                                                         e))
                     finally:
                         i += 1
                 else:
@@ -176,9 +260,12 @@ class AmsHttpRequests(object):
                     else:
                         time.sleep(retrysleep)
                         if timeout:
-                            log.warning('Retry #{0} after {1} seconds, connection timeout set to {2} seconds - {3}: {4}'.format(i, retrysleep, timeout, self.endpoint, e))
+                            log.warning(
+                                'Retry #{0} after {1} seconds, connection timeout set to {2} seconds - {3}: {4}'.format(
+                                    i, retrysleep, timeout, self.endpoint, e))
                         else:
-                            log.warning('Retry #{0} after {1} seconds - {2}: {3}'.format(i, retrysleep, self.endpoint, e))
+                            log.warning(
+                                'Retry #{0} after {1} seconds - {2}: {3}'.format(i, retrysleep, self.endpoint, e))
                 finally:
                     i += 1
 
@@ -191,14 +278,28 @@ class AmsHttpRequests(object):
         decoded = None
         try:
             # the get request based on requests.
+
+            # populate all requests with the x-api-key header
+            # except the authn mapping call
+            if route_name != "auth_x509":
+                # if there is no defined headers dict in the reqkwargs, introduce it
+                if "headers" not in reqkwargs:
+                    headers = {
+                        "x-api-key": self.token
+                    }
+                    reqkwargs["headers"] = headers
+                else:
+                    # if the there are already other headers defined, just append the x-api-key one
+                    reqkwargs["headers"]["x-api-key"] = self.token
+
             reqmethod = getattr(requests, m)
             r = reqmethod(url, data=body, **reqkwargs)
 
             content = r.content
             status_code = r.status_code
 
-            if (content and sys.version_info < (3, 6, ) and isinstance(content,
-                                                                       bytes)):
+            if (content and sys.version_info < (3, 6,) and isinstance(content,
+                                                                      bytes)):
                 content = content.decode()
 
             if status_code == 200:
@@ -246,7 +347,7 @@ class AmsHttpRequests(object):
     def do_get(self, url, route_name, **reqkwargs):
         """Method supports all the GET requests.
 
-           Used for (topics, subscriptions, messages).
+           Used for (topics, subscriptions, users, messages).
 
            Args:
                url: str. The final messaging service endpoint
@@ -277,7 +378,7 @@ class AmsHttpRequests(object):
                           python-requests library call.
         """
         # try to send a PUT request to the messaging service.
-        # if a connection problem araises a Connection error exception is raised.
+        # if a connection problem arises a Connection error exception is raised.
         try:
             return self._retry_make_request(url, body=body,
                                             route_name=route_name, **reqkwargs)
@@ -288,7 +389,7 @@ class AmsHttpRequests(object):
                 retrybackoff=None, **reqkwargs):
         """Method supports all the POST requests.
 
-           Used for (topics, subscriptions, messages).
+           Used for (topics, subscriptions, users, messages).
 
            Args:
                url: str. The final messaging service endpoint
@@ -345,81 +446,16 @@ class ArgoMessagingService(AmsHttpRequests):
        Class abstract Argo Messaging Service by covering all available HTTP API
        calls that are wrapped in series of methods.
     """
+
     def __init__(self, endpoint, token="", project="", cert="", key="", authn_port=8443):
-        super(ArgoMessagingService, self).__init__(endpoint)
-        self.authn_port = authn_port
-        self.token = ""
-        self.endpoint = endpoint
+        super(ArgoMessagingService, self).__init__(endpoint, authn_port, token, cert, key)
         self.project = project
-        self.assign_token(token, cert, key)
         self.pullopts = {"maxMessages": "1",
                          "returnImmediately": "false"}
         # Containers for topic and subscription objects
         self.topics = OrderedDict()
         self.subs = OrderedDict()
 
-    def assign_token(self, token, cert, key):
-        """Assign a token to the ams object
-
-           Args:
-               token(str): a valid ams token
-               cert(str): a path to a valid certificate file
-               key(str): a path to the associated key file for the provided certificate
-        """
-
-        # check if a token has been provided
-        if token != "":
-            self.token = token
-            return
-
-        try:
-            # otherwise use the provided certificate to retrieve it
-            self.token = self.auth_via_cert(cert, key)
-        except AmsServiceException as e:
-            # if the request send to authn didn't contain an x509 cert, that means that there was also no token provided
-            # when initializing the ArgoMessagingService object, since we only try to authenticate through authn
-            # when no token was provided
-
-            if e.msg == 'While trying the [auth_x509]: No certificate provided.':
-                refined_msg = "No certificate provided. No token provided."
-                errormsg = self._error_dict(refined_msg, e.code)
-                raise AmsServiceException(json=errormsg, request="auth_x509")
-            raise e
-
-    def auth_via_cert(self, cert, key, **reqkwargs):
-        """Retrieve an ams token based on the provided certificate
-
-            Args:
-                cert(str): a path to a valid certificate file
-                key(str): a path to the associated key file for the provided certificate
-
-           Kwargs:
-               reqkwargs: keyword argument that will be passed to underlying
-                          python-requests library call.
-        """
-        if cert == "" and key == "":
-            errord = self._error_dict("No certificate provided.", 400)
-            raise AmsServiceException(json=errord, request="auth_x509")
-
-        # create the certificate tuple needed by the requests library
-        reqkwargs = {"cert": (cert, key)}
-
-        route = self.routes["auth_x509"]
-
-        # Compose url
-        url = route[1].format(self.endpoint, self.authn_port)
-        method = getattr(self, 'do_{0}'.format(route[0]))
-
-        try:
-            r = method(url, "auth_x509", **reqkwargs)
-            # if the `token` field was not found in the response, raise an error
-            if "token" not in r:
-                errord = self._error_dict("Token was not found in the response. Response: " + str(r), 500)
-                raise AmsServiceException(json=errord, request="auth_x509")
-            return r["token"]
-        except (AmsServiceException, AmsConnectionException) as e:
-            raise e
-
     def _create_sub_obj(self, s, topic):
         self.subs.update({s['name']: AmsSubscription(s['name'], topic,
                                                      s['pushConfig'],
@@ -446,11 +482,11 @@ class ArgoMessagingService(AmsHttpRequests):
                           python-requests library call.
         """
 
-        topicobj = self.get_topic(topic, retobj=True)
+        topicobj = self.get_topic(topic, retobj=True, **reqkwargs)
 
         route = self.routes["topic_getacl"]
         # Compose url
-        url = route[1].format(self.endpoint, self.token, self.project, topic)
+        url = route[1].format(self.endpoint, self.project, topic)
         method = getattr(self, 'do_{0}'.format(route[0]))
 
         r = method(url, "topic_getacl", **reqkwargs)
@@ -475,11 +511,11 @@ class ArgoMessagingService(AmsHttpRequests):
                           python-requests library call.
         """
 
-        topicobj = self.get_topic(topic, retobj=True)
+        topicobj = self.get_topic(topic, retobj=True, **reqkwargs)
 
         route = self.routes["topic_modifyacl"]
         # Compose url
-        url = route[1].format(self.endpoint, self.token, self.project, topic)
+        url = route[1].format(self.endpoint, self.project, topic)
         method = getattr(self, 'do_{0}'.format(route[0]))
 
         r = None
@@ -506,11 +542,11 @@ class ArgoMessagingService(AmsHttpRequests):
                           python-requests library call.
         """
 
-        subobj = self.get_sub(sub, retobj=True)
+        subobj = self.get_sub(sub, retobj=True, **reqkwargs)
 
         route = self.routes["sub_getacl"]
         # Compose url
-        url = route[1].format(self.endpoint, self.token, self.project, sub)
+        url = route[1].format(self.endpoint, self.project, sub)
         method = getattr(self, 'do_{0}'.format(route[0]))
 
         r = method(url, "sub_getacl", **reqkwargs)
@@ -536,7 +572,7 @@ class ArgoMessagingService(AmsHttpRequests):
         route = self.routes["sub_offsets"]
 
         # Compose url
-        url = route[1].format(self.endpoint, self.token, self.project, sub)
+        url = route[1].format(self.endpoint, self.project, sub)
         method = getattr(self, 'do_{0}'.format(route[0]))
         r = method(url, "sub_offsets", **reqkwargs)
         try:
@@ -570,7 +606,7 @@ class ArgoMessagingService(AmsHttpRequests):
                 time_in_string = timestamp.strftime("%Y-%m-%d %H:%M:%S.000Z")
 
         # Compose url
-        url = route[1].format(self.endpoint, self.token, self.project, sub, time_in_string)
+        url = route[1].format(self.endpoint, self.project, sub, time_in_string)
 
         try:
             r = method(url, "sub_timeToOffset", **reqkwargs)
@@ -596,7 +632,7 @@ class ArgoMessagingService(AmsHttpRequests):
             move_to = int(move_to)
 
         # Compose url
-        url = route[1].format(self.endpoint, self.token, self.project, sub)
+        url = route[1].format(self.endpoint, self.project, sub)
 
         # Request body
         data = {"offset": move_to}
@@ -618,11 +654,11 @@ class ArgoMessagingService(AmsHttpRequests):
                           python-requests library call.
         """
 
-        subobj = self.get_sub(sub, retobj=True)
+        subobj = self.get_sub(sub, retobj=True, **reqkwargs)
 
         route = self.routes["sub_modifyacl"]
         # Compose url
-        url = route[1].format(self.endpoint, self.token, self.project, sub)
+        url = route[1].format(self.endpoint, self.project, sub)
         method = getattr(self, 'do_{0}'.format(route[0]))
 
         r = None
@@ -658,7 +694,7 @@ class ArgoMessagingService(AmsHttpRequests):
         msg_body = json.dumps(push_dict)
         route = self.routes["sub_pushconfig"]
         # Compose url
-        url = route[1].format(self.endpoint, self.token, self.project, "", sub)
+        url = route[1].format(self.endpoint, self.project, sub)
         method = getattr(self, 'do_{0}'.format(route[0]))
         p = method(url, msg_body, "sub_pushconfig", **reqkwargs)
 
@@ -711,7 +747,7 @@ class ArgoMessagingService(AmsHttpRequests):
         """
         route = self.routes["topic_list"]
         # Compose url
-        url = route[1].format(self.endpoint, self.token, self.project)
+        url = route[1].format(self.endpoint, self.project)
         method = getattr(self, 'do_{0}'.format(route[0]))
 
         r = method(url, "topic_list", **reqkwargs)
@@ -755,7 +791,7 @@ class ArgoMessagingService(AmsHttpRequests):
         """
         route = self.routes["topic_get"]
         # Compose url
-        url = route[1].format(self.endpoint, self.token, self.project, topic)
+        url = route[1].format(self.endpoint, self.project, topic)
         method = getattr(self, 'do_{0}'.format(route[0]))
 
         r = method(url, "topic_get", **reqkwargs)
@@ -804,7 +840,7 @@ class ArgoMessagingService(AmsHttpRequests):
 
         route = self.routes["topic_publish"]
         # Compose url
-        url = route[1].format(self.endpoint, self.token, self.project, topic)
+        url = route[1].format(self.endpoint, self.project, topic)
         method = getattr(self, 'do_{0}'.format(route[0]))
 
         return method(url, msg_body, "topic_publish", retry=retry,
@@ -820,7 +856,7 @@ class ArgoMessagingService(AmsHttpRequests):
         """
         route = self.routes["sub_list"]
         # Compose url
-        url = route[1].format(self.endpoint, self.token, self.project)
+        url = route[1].format(self.endpoint, self.project)
         method = getattr(self, 'do_{0}'.format(route[0]))
 
         r = method(url, "sub_list", **reqkwargs)
@@ -847,7 +883,7 @@ class ArgoMessagingService(AmsHttpRequests):
         """
         route = self.routes["sub_get"]
         # Compose url
-        url = route[1].format(self.endpoint, self.token, self.project, "", sub)
+        url = route[1].format(self.endpoint, self.project, sub)
         method = getattr(self, 'do_{0}'.format(route[0]))
 
         r = method(url, "sub_get", **reqkwargs)
@@ -906,7 +942,7 @@ class ArgoMessagingService(AmsHttpRequests):
 
         route = self.routes["sub_pull"]
         # Compose url
-        url = route[1].format(self.endpoint, self.token, self.project, "", sub)
+        url = route[1].format(self.endpoint, self.project, sub)
         method = getattr(self, 'do_{0}'.format(route[0]))
         r = method(url, msg_body, "sub_pull", retry=retry,
                    retrysleep=retrysleep, retrybackoff=retrybackoff,
@@ -937,7 +973,7 @@ class ArgoMessagingService(AmsHttpRequests):
 
         route = self.routes["sub_ack"]
         # Compose url
-        url = route[1].format(self.endpoint, self.token, self.project, "", sub)
+        url = route[1].format(self.endpoint, self.project, sub)
         method = getattr(self, 'do_{0}'.format(route[0]))
         method(url, msg_body, "sub_ack", **reqkwargs)
 
@@ -1017,6 +1053,103 @@ class ArgoMessagingService(AmsHttpRequests):
         """
         return self.pullopts[key]
 
+    def create_user(self, user, **reqkwargs):
+
+        """This function creates a new user with a POST request
+
+           Args:
+               user: AmsUser. The user to be created.
+               reqkwargs: keyword argument that will be passed to underlying
+               python-requests library call.
+           Return:
+               object (AmsUser)
+        """
+
+        if not isinstance(user, AmsUser):
+            raise ValueError("user has to be of type AmsUser")
+
+        try:
+            route = self.routes["user_create"]
+            url = route[1].format(self.endpoint, user.name)
+            method = getattr(self, 'do_{0}'.format(route[0]))
+            r = method(url, user.to_json(), "user_create", **reqkwargs)
+            return AmsUser().load_from_dict(r)
+        except AmsException as e:
+            raise e
+
+    def get_user(self, name, **reqkwargs):
+        """
+        Retrieves the respective user using the provided username with a GET request
+
+        :param name: (str) the username of the user to be retrieved
+        :param reqkwargs:  keyword arguments that will be passed to underlying
+               python-requests library call.
+        :return: (AmsUser) the ams user
+        """
+        try:
+            route = self.routes["user_get"]
+            url = route[1].format(self.endpoint, name)
+            method = getattr(self, 'do_{0}'.format(route[0]))
+            r = method(url, "user_get", **reqkwargs)
+            return AmsUser().load_from_dict(r)
+        except AmsException as e:
+            raise e
+
+    def add_project_member(self, username, project=None, roles=None, **reqkwargs):
+        """
+        Assigns an existing user to the provided project with a POST request
+
+        :param (str) project: the name of the project.If no
+        project is supplied, the declared global project will be used instead
+        :param (str) username: the name of user
+        :param (str[]) roles: project roles for the user
+        :param reqkwargs:  keyword argument that will be passed to underlying
+                          python-requests library call.
+        :return: (AmsUser) the assigned user object
+        """
+
+        if roles is None or not isinstance(roles, list):
+            roles = []
+
+        if project is None:
+            project = self.project
+
+        body = {
+                "roles": roles
+        }
+
+        try:
+            route = self.routes["project_add_member"]
+            url = route[1].format(self.endpoint, project, username)
+            method = getattr(self, 'do_{0}'.format(route[0]))
+            r = method(url, json.dumps(body), "project_add_member", **reqkwargs)
+            return AmsUser().load_from_dict(r)
+        except AmsException as e:
+            raise e
+
+    def get_project_member(self, username, project=None, **reqkwargs):
+        """
+        Retrieves the respective project member using the provided username with a GET request
+
+        :param username: (str) the username of the user to be retrieved
+        :param project: (str) the name of the project the user belongs to.If no
+        project is supplied, the declared global project will be used instead
+        :param reqkwargs:  keyword arguments that will be passed to underlying
+               python-requests library call.
+        :return: (AmsUser) the ams user
+        """
+        try:
+            if project is None:
+                project = self.project
+
+            route = self.routes["project_get_member"]
+            url = route[1].format(self.endpoint, project, username)
+            method = getattr(self, 'do_{0}'.format(route[0]))
+            r = method(url, "project_get_member", **reqkwargs)
+            return AmsUser().load_from_dict(r)
+        except AmsException as e:
+            raise e
+
     def create_sub(self, sub, topic, ackdeadline=10, push_endpoint=None,
                    retry_policy_type='linear', retry_policy_period=300, retobj=False, **reqkwargs):
         """This function creates a new subscription in a project with a PUT request
@@ -1044,7 +1177,7 @@ class ArgoMessagingService(AmsHttpRequests):
 
         route = self.routes["sub_create"]
         # Compose url
-        url = route[1].format(self.endpoint, self.token, self.project, "", sub)
+        url = route[1].format(self.endpoint, self.project, sub)
         method = getattr(self, 'do_{0}'.format(route[0]))
 
         r = method(url, msg_body, "sub_create", **reqkwargs)
@@ -1073,7 +1206,7 @@ class ArgoMessagingService(AmsHttpRequests):
         """
         route = self.routes["sub_delete"]
         # Compose url
-        url = route[1].format(self.endpoint, self.token, self.project, "", sub)
+        url = route[1].format(self.endpoint, self.project, sub)
         method = getattr(self, 'do_{0}'.format(route[0]))
 
         r = method(url, "sub_delete", **reqkwargs)
@@ -1119,7 +1252,7 @@ class ArgoMessagingService(AmsHttpRequests):
         """
         route = self.routes["topic_create"]
         # Compose url
-        url = route[1].format(self.endpoint, self.token, self.project, topic)
+        url = route[1].format(self.endpoint, self.project, topic)
         method = getattr(self, 'do_{0}'.format(route[0]))
 
         r = method(url, '', "topic_create", **reqkwargs)
@@ -1142,7 +1275,7 @@ class ArgoMessagingService(AmsHttpRequests):
         """
         route = self.routes["topic_delete"]
         # Compose url
-        url = route[1].format(self.endpoint, self.token, self.project, topic)
+        url = route[1].format(self.endpoint, self.project, topic)
         method = getattr(self, 'do_{0}'.format(route[0]))
 
         r = method(url, "topic_delete", **reqkwargs)
